from __future__ import print_function, division

# Extensive tests of the similarity search functions
import unittest2
import sys
import random
import contextlib

import chemfp
from chemfp import search
from chemfp import bitops
from chemfp import fps_search

import support

FULL_TEST = False

def _get_flag():
    import sys
    for (option, retval) in (
            ("--fast", "fast"),
            ("--full", "full")):
        try:
            i = sys.argv.index(option)
        except ValueError:
            continue
        del sys.argv[i]
        return retval
    return None


NUM_BITS = 128

if __name__ == "__main__":
    flag = _get_flag()
    if flag in ("fast", None):
        FULL_TEST = False
    elif flag == "full":
        FULL_TEST = True
        NUM_BITS = 1024
    else:
        raise AssertionError(flag)


_everything_arena = None
_everything_slow_arena = None
_everything_fps = None
def get_everything_data():
    global _everything_arena, _everything_fps, _everything_slow_arena
    if _everything_arena is None:
        fps = []
        bitlist = []
        for i in range(NUM_BITS+1):
            fps.append((str(i), bitops.byte_from_bitlist(bitlist, NUM_BITS)))
            bitlist.append(i)
        _everything_fps = fps
        _everything_arena = chemfp.load_fingerprints(fps, chemfp.Metadata(num_bits=NUM_BITS))
        _everything_slow_arena = _everything_arena.copy()
        _everything_slow_arena.popcount_indices = b""
    return _everything_fps, _everything_arena, _everything_slow_arena

# Code used to make "everything.fps"
## def make_everything_fps():
##     with chemfp.open_fingerprint_writer("everything.fps") as writer:
##         for id, fp in get_everything_data()[1]:
##             writer.write_fingerprint(id, fp)
## make_everything_fps()

EVERYTHING_FPS = support.fullpath("everything.fps")

def get_everything_arena():
    return get_everything_data()[1]

def get_slow_everything_arena():
    slow_arena = get_everything_data()[2]
    assert slow_arena.popcount_indices == b""
    return slow_arena

def get_both_arenas():
    return get_everything_data()[1], get_everything_data()[2]

def _get_check_bits():
    return [x for x in _raw_get_check_bits() if x <= NUM_BITS]

def _raw_get_check_bits():
    if FULL_TEST:
        #return range(arena.num_bits)
        return [0, 1, 2,
                77, 78, 79,         #   1024 / 13
                203, 204, 205,      #   1024 / 5
                253, 255, 256,      #   1024 / 4
                341, 342,           #   1024 / 3
                510, 511, 512, 513, #   1024 / 2
                682, 683,           # 2*1024 / 3
                1022, 1023,         #   1024
            ]

    else:
        return [0, 1, 2, 511, 512, 513, 1022, 1023]

interesting_thresholds = (
        0.0,
        sys.float_info.min,

        0.10373443983402489,  # 5/48.2  (1 ulp less than the next value)
        0.1037344398340249,   # 50/482  -- caused problems during initial testing
        0.10373443983402492,  # (1 ulp greater than the previous)

        0.39999999999999997,  # (1 ulp less than the next value)
        0.4,                  # 4/10
        0.4000000000000001,   # (1 ulp greater than the previous)

        0.49999999999999994,  # (1 ulp less than the next value)
        0.5,                  # 1/2
        0.5000000000000001,   # (1 ulp greater than the previous)

        0.7692307692307692,   # 1/1.3  (1 ulp less than the next value)
        0.7692307692307693,   # 10/13
        0.7692307692307694,   # (1 ulp greater than the previous)

        0.9999999999999999,   # 1 ulp less than 1.0
        1.0,
        )
assert len(set(interesting_thresholds)) == len(interesting_thresholds)

class BaseSearch(object):
    def test_empty_query(self):
        arena = self.arena
        # count
        fp = bitops.byte_from_bitlist([], NUM_BITS)
        self.assertEqual(self.count_hits_fp(fp, arena, 0.00001), 0)
        self.assertEqual(self.count_hits_fp(fp, arena, 0.00000), len(arena))
        
        # threshold
        self.assertEqual(len(self.threshold_search_fp(fp, arena, 0.00001)), 0)
        self.assertEqual(len(self.threshold_search_fp(fp, arena, 0.00000)), len(arena))

        # threshold
        self.assertEqual(len(self.knearest_search_fp(
            fp, arena, k=3, threshold=0.00001)), 0)
        self.assertEqual(len(self.knearest_search_fp(
            fp, arena, k=3, threshold=0.00000)), 3)
        self.assertEqual(len(self.knearest_search_fp(
            fp, arena, k=len(arena), threshold=0.00000)), len(arena))

        
    def test_single_bit_scan_count(self):
        arena = self.arena
        check_bits = _get_check_bits()
        
        for bitno in check_bits:
            fp = bitops.byte_from_bitlist([bitno], NUM_BITS)
            expected_scores = [(i>bitno)/(i+(i<=bitno)) for i in range(NUM_BITS+1)]
            expected_scores.sort(reverse=True)
            # Check at the threshold borders
            for threshold in sorted(set(expected_scores)):
                # Are the theshold counts correct?
                found_count = self.count_hits_fp(fp, arena, threshold)
                # We go from lowest threshold to highest, so keep reducing the list
                expected_scores = [score for score in expected_scores if score >= threshold]
                expected_count = len(expected_scores)
                self.assertEqual(found_count, expected_count, (bitno, threshold))


    def test_single_bit_threshold(self):
        arena = self.arena
        check_bits = _get_check_bits()
        
        for bitno in check_bits:
            fp = bitops.byte_from_bitlist([bitno], NUM_BITS)
            expected_scores = [(i>bitno)/(i+(i<=bitno)) for i in range(NUM_BITS+1)]
            expected_scores.sort(reverse=True)
            # Check at the threshold borders
            for threshold in sorted(set(expected_scores)):
                # We go from lowest threshold to highest, so keep reducing the list
                expected_scores = [score for score in expected_scores if score >= threshold]
        
                # Are the theshold scores correct?
                found_hits = self.threshold_search_fp(fp, arena, threshold)
                found_hits.reorder("decreasing-score")
                found_scores = list(found_hits.get_scores())
                self.assertEqual(found_scores, expected_scores, (bitno, threshold))

    def test_single_bit_knearest(self):
        arena = self.arena
        check_bits = _get_check_bits()
        
        for bitno in check_bits:
            fp = bitops.byte_from_bitlist([bitno], NUM_BITS)
            expected_scores = [(i>bitno)/(i+(i<=bitno)) for i in range(NUM_BITS+1)]
            expected_scores.sort(reverse=True)
            # Check at the threshold borders
            for threshold in sorted(set(expected_scores)):
                # We go from lowest threshold to highest, so keep reducing the list
                expected_scores = [score for score in expected_scores if score >= threshold]
                        
                ## Are the k-nearest searches correct?
                found_hits = self.knearest_search_fp(fp, arena, len(arena), threshold)
                found_scores = list(found_hits.get_scores())
                self.assertEqual(found_scores, expected_scores, (bitno, threshold))
                
                ## found_hits = search.knearest_tanimoto_search_fp(fp, arena, 5, threshold)
                ## found_scores = list(found_hits.get_scores())
                ## self.assertEqual(found_scores, expected_scores[:5], (bitno, threshold))
                ## self.assertEqual(len(found_hits), 5, (bitno, threshold))

    def test_count_arena(self):
        exact_hits = compute_exact_scaled_tversky_scores(10, 10)
        for threshold in interesting_thresholds:
            counts = self.count_hits_arena(get_one_bit_arena(), self.arena, threshold)
            expected_counts = [sum(1 for hit in hits if hit[1] >= threshold) for hits in exact_hits]
            self.assertEqual(list(counts), expected_counts, (threshold,))
        
    def test_threshold_arena(self):
        # This is the same code path as the fp x arena fingerprints
        # so I am more cursory
        arena = self.arena
        assert len(arena) == NUM_BITS+1
        query_arena = chemfp.load_fingerprints(
            [(str(bitno), bitops.byte_from_bitlist([bitno], NUM_BITS))
                   for bitno in range(NUM_BITS)], arena.metadata)
        
        all_expected_hits = []
        for bitno in range(NUM_BITS):
            expected_hits = [(str(i), (i>bitno)/(i+(i<=bitno))) for i in range(NUM_BITS+1)]
            expected_hits.sort(reverse=True, key=lambda x: x[1])
            all_expected_hits.append(expected_hits)

        for threshold in interesting_thresholds:
            all_hits = self.threshold_search_arena(
                query_arena, arena, threshold=threshold)
            all_hits.reorder_all("decreasing-score")
            for query_i, hits in enumerate(all_hits):
                expected = [x for x in all_expected_hits[query_i] if x[1] >= threshold]
                self.assertEqual(sorted(hits.get_ids_and_scores()), sorted(expected), (query_i, threshold))


                
    def test_knearest_arena(self):
        # This is the same code path as the fp x arena fingerprints
        # so I am more cursory
        arena = self.arena
        assert len(arena) == NUM_BITS+1
        query_arena = chemfp.load_fingerprints(
            [(str(bitno), bitops.byte_from_bitlist([bitno], NUM_BITS))
                for bitno in range(NUM_BITS)], arena.metadata)
        assert len(query_arena) == NUM_BITS
        
        all_expected_scores = []
        for bitno in range(NUM_BITS):
            expected_scores = [(i>bitno)/(i+(i<=bitno)) for i in range(NUM_BITS+1)]
            expected_scores.sort(reverse=True)
            all_expected_scores.append(expected_scores)

        for threshold in interesting_thresholds:

            all_hits = self.knearest_search_arena(
                query_arena, arena, k=len(arena), threshold=threshold)
            all_hits.reorder_all("decreasing-score")
            for query_i, hits in enumerate(all_hits):
                expected = [score for score in all_expected_scores[query_i] if score >= threshold]
                self.assertEqual(len(hits), len(expected), (query_i, threshold))
                self.assertEqual(list(hits.get_scores()), expected, (query_i, threshold))

            all_hits = self.knearest_search_arena(
                query_arena, arena, k=5, threshold=threshold)
            all_hits.reorder_all("decreasing-score")
            for query_i, hits in enumerate(all_hits):
                expected = [score for score in all_expected_scores[query_i] if score >= threshold]
                if len(expected) > 5:
                    expected = expected[:5]
                self.assertEqual(list(hits.get_scores()), expected, (query_i, threshold))


class Symmetric(object):
    def test_count_hits_symmetric(self):
        # At this point I trust the fp x arena search.
        # Use that to test the symmetric code
        arena = get_everything_arena()
        for threshold in interesting_thresholds:
            
            all_counts = self.count_hits_symmetric(arena, threshold)
            for query_i, (query_id, query_fp) in enumerate(arena):
                expected_count = self.count_hits_fp(query_fp, arena, threshold)
                if threshold == 0.0:
                    expected_count = len(arena)-1
                elif query_i == 0:
                    # 0/0 case
                    assert expected_count == 0, (expected_count, threshold)
                else:
                    expected_count -= 1
                
                self.assertEqual(all_counts[query_i], expected_count, (query_i, query_id, threshold, query_fp))


    def test_threshold_search_symmetric(self):
        # At this point I trust the fp x arena search.
        # Use that to test the symmetric code
        arena = get_everything_arena()
        for threshold in interesting_thresholds:
            
            all_hits = self.threshold_search_symmetric(arena, threshold)
            all_hits.reorder_all("decreasing-score")
            for query_i, (query_id, query_fp) in enumerate(arena):
                fp_hits = self.threshold_search_fp(query_fp, arena, threshold)
                fp_hits.reorder("decreasing-score")
                expected_hits = [x for x in fp_hits.get_ids_and_scores() if x[0] != query_id]
                self.assertEqual(all_hits[query_i].get_ids_and_scores(),
                                 expected_hits,
                                 (query_i, threshold))
                        
    def test_knearest_search_symmetric(self):
        # At this point I trust the fp x arena search.
        # Use that to test the k-nearest symmetric code
        arena = get_everything_arena()
        for threshold in interesting_thresholds:
            all_hits = self.knearest_search_symmetric(arena, len(arena), threshold)
            if threshold == 0.0:
                for query_i, hits in enumerate(all_hits):
                    if len(hits) != len(arena)-1:
                        print("oops", hits.get_indices_and_scores())
                    self.assertEqual(len(hits), len(arena)-1, (query_i, threshold))
                
            #print("QQQ", all_hits[0].get_indices_and_scores())
            for query_i, (query_id, query_fp) in enumerate(arena):
                fp_hits = self.knearest_search_fp(query_fp, arena, len(arena), threshold)
                expected_hits = [x for x in fp_hits.get_ids_and_scores() if x[0] != query_id]
                self.assertEqual(all_hits[query_i].get_ids_and_scores()[:10], expected_hits[:10], (query_i, query_id, threshold))
                self.assertEqual(all_hits[query_i].get_ids_and_scores(), expected_hits, (query_i, query_id, threshold))

            all_hits = self.knearest_search_symmetric(arena, 5, threshold)
            for query_i, (query_id, query_fp) in enumerate(arena):
                hits = self.knearest_search_fp(query_fp, arena, 6, threshold)
                self.assertEqual(all_hits[query_i].get_ids_and_scores(),
                                 [x for x in hits.get_ids_and_scores() if x[0] != query_id])

class SingleThreaded(object):
    def setUp(self):
        self._num_threads = chemfp.get_num_threads()
        chemfp.set_num_threads(0)
        
    def tearDown(self):
        chemfp.set_num_threads(self._num_threads)

class MultiThreaded(object):
    def setUp(self):
        self._num_threads = chemfp.get_num_threads()
        chemfp.set_num_threads(4)
        
    def tearDown(self):
        chemfp.set_num_threads(self._num_threads)
        
class UseIndexedArena(Symmetric):
    def setUp(self):
        self.arena = get_everything_arena()

class UseUnindexedArena(object):
    def setUp(self):
        self.arena = get_slow_everything_arena()

class Tanimoto(object):        
    count_hits_fp = staticmethod(search.count_tanimoto_hits_fp)
    threshold_search_fp = staticmethod(search.threshold_tanimoto_search_fp)
    knearest_search_fp = staticmethod(search.knearest_tanimoto_search_fp)
    count_hits_arena = staticmethod(search.count_tanimoto_hits_arena)
    threshold_search_arena = staticmethod(search.threshold_tanimoto_search_arena)
    knearest_search_arena = staticmethod(search.knearest_tanimoto_search_arena)
    count_hits_symmetric = staticmethod(search.count_tanimoto_hits_symmetric)
    threshold_search_symmetric = staticmethod(search.threshold_tanimoto_search_symmetric)
    knearest_search_symmetric = staticmethod(search.knearest_tanimoto_search_symmetric)

class TestTanimotoSingleThreadedIndexed(BaseSearch, Tanimoto, SingleThreaded, UseIndexedArena, unittest2.TestCase):
    def setUp(self):
        SingleThreaded.setUp(self)
        UseIndexedArena.setUp(self)
        self.assertTrue(self.arena.popcount_indices)
    
class TestTanimotoSingleThreadedUnindexed(BaseSearch, Tanimoto, SingleThreaded, UseUnindexedArena, unittest2.TestCase):
    def setUp(self):
        SingleThreaded.setUp(self)
        UseUnindexedArena.setUp(self)
        self.assertFalse(self.arena.popcount_indices)
    
class TestTanimotoMultiThreadedIndexed(BaseSearch, Tanimoto, MultiThreaded, UseIndexedArena, unittest2.TestCase):
    def setUp(self):
        MultiThreaded.setUp(self)
        UseIndexedArena.setUp(self)
        self.assertTrue(self.arena.popcount_indices)
    
class TestTanimotoMultiThreadedUnindexed(BaseSearch, Tanimoto, MultiThreaded, UseUnindexedArena, unittest2.TestCase):
    def setUp(self):
        MultiThreaded.setUp(self)
        UseUnindexedArena.setUp(self)
        self.assertFalse(self.arena.popcount_indices)
    
        

# For the Tversky NxM queries, set up a query arena with each on-bit set

_one_bit_arena = None
def get_one_bit_arena():
    global _one_bit_arena
    if _one_bit_arena is None:
        metadata = chemfp.Metadata(num_bits=NUM_BITS)
        _one_bit_arena = chemfp.load_fingerprints(
            ((str(bitno), bitops.byte_from_bitlist([bitno], NUM_BITS))
                  for bitno in range(NUM_BITS)),
            metadata)
    return _one_bit_arena

def compute_exact_scaled_tversky_scores(scaled_alpha, scaled_beta):
    queries = get_one_bit_arena()
    
    all_hits = []
    for query_bitno in range(len(queries)):
        hits = []
        for target_num_bits in range(NUM_BITS+1):
            if target_num_bits == 0:
                c = 0
            else:
                if query_bitno < target_num_bits:
                    c = 1
                else:
                    c = 0
            if c == 0:
                score = 0.0
            else:
                score = 10*c / (scaled_alpha * (1-c) + scaled_beta * (target_num_bits-c) + 10*c)
            hits.append((str(target_num_bits), score))
        all_hits.append(hits)
    return all_hits


def compute_exact_scaled_tversky_scores_symmetric(alpha, beta):
    all_hits = []
    str_ids = [str(x) for x in range(NUM_BITS+1)]
    for num_bits1 in range(NUM_BITS+1):
        hits = []
        for num_bits2 in range(num_bits1):
            A = num_bits1 - num_bits2
            B = 0
            c = num_bits2
            score = 10*c / (alpha * A + beta * B + 10*c)
            hits.append( (str_ids[num_bits2], score) )

        for num_bits2 in range(num_bits1+1, NUM_BITS+1):
            A = 0
            B = num_bits2 - num_bits1
            c = num_bits1
            score = 10*c / (alpha * A + beta * B + 10*c)
            hits.append( (str_ids[num_bits2], score) )
        #scores.sort()
        all_hits.append(hits)
    return all_hits


#### The file-based tests

class FakeArena(object):
    def __init__(self):
        arena = get_everything_arena()
        self._len = len(arena)
        self.metadata = arena.metadata
    def __len__(self):
        return self._len

class FPSTanimoto(object):
    def setUp(self):
        self.arena = FakeArena()

    @contextlib.contextmanager
    def _get_targets(self):
        reader = chemfp.open(EVERYTHING_FPS)
        try:
            yield reader
        finally:
            reader.close()
        
    def count_hits_fp(self, fp, targets, threshold):
        with self._get_targets() as targets:
            return fps_search.count_tanimoto_hits_fp(fp, targets, threshold)
        
    def threshold_search_fp(self, fp, targets, threshold):
        with self._get_targets() as targets:
            return fps_search.threshold_tanimoto_search_fp(fp, targets, threshold)

    def knearest_search_fp(self, fp, targets, k, threshold):
        with self._get_targets() as targets:
            return fps_search.knearest_tanimoto_search_fp(fp, targets, k, threshold)
    
    def count_hits_arena(self, queries, targets, threshold):
        with self._get_targets() as targets:
            return fps_search.count_tanimoto_hits_arena(queries, targets, threshold)
        
    def threshold_search_arena(self, queries, targets, threshold):
        with self._get_targets() as targets:
            return fps_search.threshold_tanimoto_search_arena(queries, targets, threshold)
        
    def knearest_search_arena(self, queries, targets, k, threshold):
        with self._get_targets() as targets:
            return fps_search.knearest_tanimoto_search_arena(queries, targets, k, threshold)
        
class TestFPSTanimotoSearch(BaseSearch, FPSTanimoto, unittest2.TestCase):
    pass


class TestArenaMemoryLeak(unittest2.TestCase):
    def test_arena_leak(self):
        arena = get_everything_arena()
        subarena = arena.copy(indices=[55, 66, 77, 88, 99])
        arena_refcount = sys.getrefcount(arena.arena)
        arena_indices_refcount = sys.getrefcount(arena.popcount_indices)
        subarena_refcount = sys.getrefcount(subarena.arena)
        subarena_indices_refcount = sys.getrefcount(subarena.popcount_indices)
        def check_counts():
            self.assertEqual(sys.getrefcount(arena.arena), arena_refcount)
            self.assertEqual(sys.getrefcount(arena.popcount_indices), arena_indices_refcount)
            self.assertEqual(sys.getrefcount(subarena.arena), subarena_refcount)
            self.assertEqual(sys.getrefcount(subarena.popcount_indices), subarena_indices_refcount)

        # Regression tests for a memory leak caught during the beta relase of v3.0.
        search.knearest_tanimoto_search_arena(arena, subarena)
        check_counts()
        search.knearest_tanimoto_search_arena(arena, subarena, k=8, threshold=0.2)
        check_counts()

        search.threshold_tanimoto_search_arena(arena, subarena)
        check_counts()
        search.threshold_tanimoto_search_arena(arena, subarena, threshold=0.4)
        check_counts()
        
        search.count_tanimoto_hits_arena(arena, subarena)
        check_counts()
        search.count_tanimoto_hits_arena(arena, subarena, threshold=0.3)
        check_counts()

        search.threshold_tanimoto_search_symmetric(arena)
        check_counts()
        search.threshold_tanimoto_search_symmetric(subarena)
        check_counts()

        search.count_tanimoto_hits_symmetric(arena)
        check_counts()
        search.count_tanimoto_hits_symmetric(subarena)
        check_counts()

        search.knearest_tanimoto_search_symmetric(arena)
        check_counts()
        search.knearest_tanimoto_search_symmetric(subarena)
        check_counts()
    
        # Found during v3.2 development
        search.contains_arena(subarena, arena)
        check_counts()
        
            
if __name__ == "__main__":
    flag = _get_flag()
    if flag in ("fast", None):
        FULL_TEST = False
    elif flag == "full":
        FULL_TEST = True
    else:
        raise AssertionError(flag)
    unittest2.main()



# Verify
import re
import itertools
import string

from chemfp.decoders import from_cactvs
from chemfp import types
from chemfp.bitops import hex_decode, hex_encode


# Parse terms which look like "34","82-88", and "0-4,9,23-880"

_range_pat = re.compile("(\d+)-(\d+)")
_bit_pat = re.compile("\d+")

def parse_bitset_definition(s):
    if s.endswith("=="):
        num_bits, decoded = from_cactvs(s)
        return set(bitno for (bitno, bit) in iter_bits(decoded) if bit)

    if s.startswith("0x"):
        decoded = hex_decode(s[2:])
        return set(bitno for (bitno, bit) in iter_bits(decoded) if bit)

    bits = set()
    if s == "X":
        return bits

    start = 0
    while 1:
        m = _range_pat.match(s, start)
        if m:
            # Ranges are inclusive, so "3-5" means {3, 4, 5}
            start = int(m.group(1))
            end = int(m.group(2))
            assert start <= end, (start, end)
            for bit in range(start, end+1):
                bits.add(bit)
            start = m.end()
        else:
            m = _bit_pat.match(s, start)
            if m:
                # a single bit
                bits.add(int(m.group()))
                start = m.end()
            else:
                raise ValueError("Unknown term %r position %d" % (s, start+1))
        if start == len(s):
            break
        if s[start] != ",":
            raise ValueError("Must have a ',' in %r at position %d" % (s, start+1))
        start += 1

    return bits

def parse_true(true_term, num_bits):
    if true_term.endswith("?"):
        assert "!" not in true_term
        true_bits = parse_bitset_definition(true_term[:-1])
        ignore_bits = set(range(num_bits)) - true_bits
        return true_bits, ignore_bits

    if "!" not in true_term:
        return parse_bitset_definition(true_term), set()

    left, right = true_term.split("!")
    return parse_bitset_definition(left), parse_bitset_definition(right)
    

def parse_false(false_term, num_bits, true_bits):
    if false_term == "*":
        return set(range(num_bits)) - true_bits
    return parse_bitset_definition(false_term)


_bit_offset_table = {}
for i in range(256):
    _bit_offset_table[chr(i)] = tuple(offset for offset in range(8) if i & (1<<offset))

def iter_bits(fp):
    for byteno, byte in enumerate(fp):
        base = byteno*8
        word = ord(byte)
        for j in (0,1,2,3,4,5,6,7):
            yield base+j, bool(word&(1<<j))


def get_fingerprinter(line, toolkit):
    fields = line.split()
    family = fields[0].format(TOOLKIT=toolkit)
    fields[0] = family
    return types.parse_type(" ".join(fields))

def evaluate_test_cases(fingerprinter, test_cases):
    f = open("tmp.smi", "w")
    for i, test_case in enumerate(test_cases):
        f.write("%s %d\n" % (test_case[0], i))
    f.close()

    errors = []
    tested_true = set()
    tested_false = set()

    reader = fingerprinter.read_structure_fingerprints("tmp.smi", "smi")
    for i, (test_case, fp_result) in enumerate(
                           itertools.izip_longest(test_cases, reader) ):
        assert test_case is not None, "too many results?"
        smiles, true_bits, false_bits, ignore_bits = test_case
        print "SMILES", smiles
        assert fp_result is not None, "not enough results?"
        fp, title = fp_result
        assert str(i) == title, (i, title)


        for bitno, val in iter_bits(fp):
            if bitno >= fingerprinter.num_bits:
                break
            if bitno in ignore_bits:
                continue
            if bitno in true_bits:
                if not val:
                    errors.append( (bitno, val, True) )
                else:
                    tested_true.add(bitno)
            elif bitno in false_bits:
                if val:
                    errors.append( (bitno, val, False) )
                else:
                    tested_false.add(bitno)

        if errors:
            print "ERROR: Fingerprint failure:", smiles
            print "Fingerprint", fingerprinter.get_type()
            print " bit#  got  expected"
            print " ----  ---  --------"
            for (bitno, got, expected) in errors:
                description = fingerprinter.describe(bitno)
                print " %3d    %d      %d  %s" % (bitno, got, expected, description)
            print "bit pattern", ",".join(str(bitno) for (bitno, val) in iter_bits(fp) if val)
            print "hex pattern", hex_encode(fp)
            raise SystemExit()

    all = set(range(fingerprinter.num_bits))
    print "Missing true bits:", " ".join(str(b) for b in sorted(all - tested_true))
    print "Missing false bits:", " ".join(str(b) for b in sorted(all - tested_true))
    
        

def main():
    import sys
    if len(sys.argv) > 1:
        toolkit = sys.argv[1]
        main2(toolkit)
    else:
        for toolkit in ("OpenEye", "RDKit", "OpenBabel", "Indigo"):
            main2(toolkit)

def main2(toolkit):
    fingerprinter = None

    num_bits = None
    toolkit_version = None
    
    test_cases = []
    skip = False

    def toolkit_selected(options):
        print toolkit, toolkit_version, options
        return toolkit in options or toolkit_version in options
    
    for line in open("rdmaccs.fpverify", "U"):
        line = line.strip()
        if not line:
            continue
        
        if line.startswith("#type="):
            assert fingerprinter is None
            fingerprinter = get_fingerprinter(line[6:], toolkit)
            num_bits = fingerprinter.num_bits
            toolkit_version = fingerprinter.software.split()[0]
            continue

        if line.startswith("#"):
            continue

        fields = line.split()
        
        if fields[0] == "=skip":
            if toolkit_selected(fields[1:]):
                print "skip", fields[1:]
                skip = True
            continue

        if fields[0] == "=only":
            if not toolkit_selected(fields[1:]):
                skip = True
            continue

        if fields[0].startswith("="):
            if fields[0][1:2] in string.ascii_uppercase:
                # Some sort of toolkit-specific directive. Ignore.
                continue
            raise AssertionError(line)
        
        assert fingerprinter is not None # haven't defined the fingerprint type

        if skip:
            skip = False
            continue


        if len(fields) == 1:
            raise AssertionError(line)
        if len(fields) == 2:
            smiles, true_term = fields
            false_term = "*"
        else:
            smiles, true_term, false_term = fields # must have 2 or 3 terms

        true_bits, ignore_bits = parse_true(true_term, num_bits)
        false_bits = parse_false(false_term, num_bits, true_bits)

        true_bits = true_bits - ignore_bits
        false_bits = false_bits - ignore_bits

        assert not (true_bits & false_bits), (true_bits & false_bits)

        test_cases.append( (smiles, true_bits, false_bits, ignore_bits) )

    evaluate_test_cases(fingerprinter, test_cases)

if __name__ == "__main__":
    main()



# Copyright (c) 2010-2017 Andrew Dalke Scientific, AB (Sweden)
# See the contents of "chemfp/__init__.py" for full license details.

from __future__ import absolute_import

from openeye.oechem import (
    OESubSearch, OEChemGetRelease, OEChemGetVersion, OEGraphMol, OEAndAtom,
    OENotAtom, OEIsAromaticAtom, OEIsCarbon, OEIsAromaticBond, OEAtomIsInRing, OEHasBondIdx,
    OEFindRingAtomsAndBonds, OEDetermineAromaticRingSystems, OEDetermineComponents)


from . import openeye
from . import pattern_fingerprinter
from . import types
from . import __version__ as chemfp_version
        
class HydrogenMatcher(object):
    def __init__(self, max_count):
        self.max_count = max_count

    def SingleMatch(self, mol):
        for atom in mol.GetAtoms():
            if atom.GetAtomicNum() == 1:
                return 1
            if atom.GetImplicitHCount():
                return 1
        return 0

    def Match(self, mol, flg=True):
        max_count = self.max_count
        count = 0
        for atom in mol.GetAtoms():
            if atom.GetAtomicNum() == 1:
                count += 1
            count += atom.GetImplicitHCount()
            if count > max_count:
                break
        return [0] * count

# OpenEye famously does not include SSSR functionality in OEChem.
# Search for "Smallest Set of Smallest Rings (SSSR) Considered Harmful"
# After much thought, I agree. But it makes this sort of code harder.

# That's why I only support up to max_count = 2. Then again, I know
# that this code does the right thing, while I'm not sure about the
# SSSR-based implementations.

class AromaticRings(object):
    def __init__(self, max_count):
        if max_count > 2:
            raise NotImplementedError("No support for >=3 aromatic rings")
        self.max_count = max_count
        self._single_aromatic = OESubSearch("[aR]")
        # In OpenEye SMARTS, [a;!R2] will find aromatic atoms in at least two rings
        # The following finds atoms which are members of at least two aromatic rings
        self._multiring_aromatic = OESubSearch("[a;!R2](:a)(:a):a")
        
    def SingleMatch(self, mol):
        # This is easy; if there's one aromatic atom then there's one
        # aromatic ring.
        return self._single_aromatic.SingleMatch(mol)
        
    def Match(self, mol, flg=True):
        # We're trying to find if there are two aromatic rings.
        if not self._single_aromatic.SingleMatch(mol):
            return ()
            
        if self._multiring_aromatic.SingleMatch(mol):
            # then obviously there are two aromatic rings
            return (1,2)

        # Since no aromatic atom is in two aromatic rings that means
        # the aromatic ring systems are disjoint, so this gives me the
        # number of ring systems
        num_aromatic_systems, parts = OEDetermineAromaticRingSystems(mol)
        if num_aromatic_systems >= self.max_count:
            return [0]*self.max_count
        
        assert num_aromatic_systems != 0, "there is supposed to be an aromatic ring"
        if num_aromatic_systems == 1:
            return (1,)
        raise AssertionError("Should not get here")
    
_is_hetereo_aromatic = OEAndAtom(OEAndAtom(OEIsAromaticAtom(), OENotAtom(OEIsCarbon())), OEAtomIsInRing())
class HeteroAromaticRings(object):
    def __init__(self, max_count):
        if max_count > 2:
            raise NotImplementedError("No support for >=3 hetero-aromatic rings")
        self.max_count = max_count

    def SingleMatch(self, mol):
        for atom in mol.GetAtoms(_is_hetereo_aromatic):
            return True
        return False

    def Match(self, mol, flg=True):
        # Find all the hetero-aromatic atoms
        hetero_atoms = [atom for atom in mol.GetAtoms(_is_hetereo_aromatic)]
        if len(hetero_atoms) < 2:
            # The caller just needs an iterable
            return hetero_atoms

        # There are at least two hetero-aromatic atoms.
        # Are there multiple ring systems?
        num_aromatic_systems, parts = OEDetermineAromaticRingSystems(mol)
        assert num_aromatic_systems >= 1
        # Are there hetero-atoms in different systems?
        atom_components = set(parts[atom.GetIdx()] for atom in hetero_atoms)
        if len(atom_components) > 1:
            return (1,2)

        # The answer now is "at least one". But are there two?

        # All of the hetero-aromatic atoms are in the same ring system
        # This is the best answer I could think of, and it only works
        # with the OEChem toolkit: remove one of the bonds, re-find
        # the rings, and see if there's still an aromatic hetero-atom.
        hetero_atom = hetero_atoms[0]

        for bond in hetero_atom.GetBonds(OEIsAromaticBond()):
            newmol = OEGraphMol(mol)
            newmol_bond = newmol.GetBond(OEHasBondIdx(bond.GetIdx()))
            newmol.DeleteBond(newmol_bond)
            OEFindRingAtomsAndBonds(newmol)

            for atom in newmol.GetAtoms(_is_hetereo_aromatic):
                return (1,2)

        return (1,)

class NumFragments(object):
    def __init__(self, max_count):
        pass
    def SingleMatch(self, mol):
        return mol.NumAtoms() > 0
    def Match(self, mol, flg=True):
        count, parts = OEDetermineComponents(mol)
        # parts is a list of component numbers.
        # Turn them into a set to get the unique set of component numbers
        # Sets are iterable, so I don't need to do more for the API
        return set(parts)

# Grrr. The substructure keys want up to 4 aromatic rings. The above
# code only works for up to 2. The API doesn't let me say "I can
# handle up to 2; please set the remainder to 0."
#
# XXX Well, I can change that.

def aromatic_rings(max_count):
    if max_count > 2:
        return pattern_fingerprinter.LimitedMatcher(2, AromaticRings(2))
    return AromaticRings(max_count)

def hetero_aromatic_rings(max_count):
    if max_count > 2:
        return pattern_fingerprinter.LimitedMatcher(2, HeteroAromaticRings(2))
    return HeteroAromaticRings(max_count)

_pattern_classes = {
    "<H>": HydrogenMatcher,
    "<aromatic-rings>": aromatic_rings,
    "<hetero-aromatic-rings>": hetero_aromatic_rings,
    "<fragments>": NumFragments,
    }
    

def oechem_compile_pattern(pattern, max_count):
    if pattern in _pattern_classes:
        return _pattern_classes[pattern](max_count)

    elif pattern.startswith("<"):
        raise NotImplementedError(pattern) # No other special patterns are supported
    else:
        pat = OESubSearch()
        if not pat.Init(pattern):
            raise pattern_fingerprinter.UnsupportedPatternError(
                pattern, "Uninterpretable SMARTS pattern")
        pat.SetMaxMatches(max_count)
        return pat


class OEChemPatternFingerprinter(pattern_fingerprinter.PatternFingerprinter):
    def __init__(self, patterns):
        assert patterns is not None
        super(OEChemPatternFingerprinter, self).__init__(patterns, oechem_compile_pattern)
        
    def fingerprint(self, mol):
        bytes = [0] * self.num_bytes
        for matcher, largest_count, count_info_tuple in self.matcher_definitions:
            if matcher is NotImplemented:
                continue
            #print matcher, largest_count, count_info_tuple
            if largest_count == 1:
                if matcher.SingleMatch(mol):
                    count_info = count_info_tuple[0]
                    bytes[count_info.byteno] |= count_info.bitmask
            else:
                actual_count = sum(1 for ignore in matcher.Match(mol, True)) # unique matches
                if actual_count:
                    for count_info in count_info_tuple:
                        if actual_count >= count_info.count:
                            bytes[count_info.byteno] |= count_info.bitmask
                        else:
                            break
        return "".join(map(chr, bytes))

class _CachedFingerprinters(dict):
    def __missing__(self, name):
        patterns = pattern_fingerprinter._load_named_patterns(name)
        fingerprinter = OEChemPatternFingerprinter(patterns)
        self[name] = fingerprinter
        return fingerprinter
_cached_fingerprinters = _CachedFingerprinters()
        

SOFTWARE = "OEChem/%(release)s (%(version)s) chemfp/%(chemfp)s" % dict(
    # There can be spaces in the beta releases
    release = OEChemGetRelease().replace(" ", "_"),
    version = OEChemGetVersion(),
    chemfp = chemfp_version)


# XXX Why are there two "Fingerprinter" classes?
# XX Shouldn't they be merged?

_base = openeye._base.clone(
    software = SOFTWARE)

SubstructOpenEyeFingerprinter_v1 = _base.clone(
    name = "ChemFP-Substruct-OpenEye/1",
    num_bits = 881,
    make_fingerprinter = lambda : _cached_fingerprinters["substruct"].fingerprint)
    
#    def describe(self, bitno):
#        return self._fingerprinter.describe(bitno)

RDMACCSOpenEyeFingerprinter_v1 = _base.clone(
    name = "RDMACCS-OpenEye/1",
    num_bits = 166,
    make_fingerprinter = lambda : _cached_fingerprinters["rdmaccs"].fingerprint)

RDMACCSOpenEyeFingerprinter_v2 = _base.clone(
    name = "RDMACCS-OpenEye/2",
    num_bits = 166,
    make_fingerprinter = lambda : _cached_fingerprinters["rdmaccs2"].fingerprint)



from __future__ import with_statement
import os
import shutil
import tempfile
import unittest2

import support
from chemfp.bitops import hex_contains_bit

try:
    import openbabel
    has_openbabel = True
    skip_openbabel = False
except ImportError:
    has_openbabel = False
    skip_openbabel = True

    if not support.can_skip("ob"):
        skip_openbabel = False
        import openbabel
        
if has_openbabel:
    import chemfp.openbabel
    from chemfp.commandline import ob2fps
    VERSION = chemfp.openbabel._ob_version

    runner = support.Runner(ob2fps.main)
    run = runner.run
    run_fps = runner.run_fps
    run_split = runner.run_split
    run_exit = runner.run_exit

    HAS_MACCS = chemfp.openbabel.HAS_MACCS
    MACCS_SMI = support.fullpath("maccs.smi")
else:
    HAS_MACCS = False
    runner = None

class TestFingerprintTypes(unittest2.TestCase):
    def test_unspecified(self):
        # Should give the same results as FP2
        headers, fps = run_split("", 19)
        self.assertEquals(headers["#type"], "OpenBabel-FP2/1")
        self.assertEquals(fps[0], "200206000000000402800e00040140010100014008206200000200c0082200080200500201c9804100270538000000402000a2040080c1240001c2c2004600200c200c04020800200410a0001490000200a803c018005400c80c00000000810100840000880064a0124010000000080102060142400110200a00000004800000\t9425004")
    def test_FP2(self):
        headers, fps = run_split("--FP2", 19)
        self.assertEquals(headers["#type"], "OpenBabel-FP2/1")
        self.assertEquals(fps[0], "200206000000000402800e00040140010100014008206200000200c0082200080200500201c9804100270538000000402000a2040080c1240001c2c2004600200c200c04020800200410a0001490000200a803c018005400c80c00000000810100840000880064a0124010000000080102060142400110200a00000004800000\t9425004")
    def test_FP3(self):
        headers, fps = run_split("--FP3", 19)
        self.assertEquals(headers["#type"], "OpenBabel-FP3/1")
        self.assertEquals(fps[0], "0400000000b001\t9425004")
    def test_FP4(self):
        headers, fps = run_split("--FP4", 19)
        self.assertEquals(headers["#type"], "OpenBabel-FP4/1")
        # Sigh. OpenBabel post 2.3 added stereo support. This structure is
        #   Clc1c(/C=C/C(=O)NNC(=O)Cn2nc(cc2C)C)c(F)ccc1\t9425004\n
        # FP4 bits 289 and 290 are "on" for post 2.3.0 releases
        # Those bits are:
        #  289 Cis_double_bond: */[D2]=[D2]\*
        #  290 Trans_double_bond: */[D2]=[D2]/*
        if (VERSION.startswith("2.2") or
            VERSION == "2.3.0"):
            self.assertEquals(fps[0], "1100000000000000000080000000000000010000000c9800000000000000000000000640407800\t9425004")  # old (without stereo)
        else:
            self.assertEquals(fps[0], "1100000000000000000080000000000000010000000c9800000000000000000000000640437800\t9425004")  # new (with stereo)

    @unittest2.skipUnless(HAS_MACCS, "Missing MACCS support")
    def test_MACCS(self):
        headers, fps = run_split("--MACCS", 19)
        if headers["#type"] == "OpenBabel-MACCS/1":
            # Running on a buggy 2.3.0 release or earlier
            self.assertEquals(fps[0], "800400000002080019cc40eacdec980baea378ef1b\t9425004")
        else:
            self.assertEquals(headers["#type"], "OpenBabel-MACCS/2")
            # Running on a corrected post-2.3.0 release
            self.assertEquals(fps[0], "000000000002080019c444eacd6c980baea178ef1f\t9425004")
            
    @unittest2.skipIf(HAS_MACCS, "check for missing MACCS support")
    def test_MACCS_does_not_exist(self):
        run_exit("--MACCS")


    def test_rdmaccs(self):
        headers, fps = run_split("--rdmaccs", 19)
        self.assertEquals(headers["#type"], "RDMACCS-OpenBabel/2")
        self.assertEquals(fps[0], "000000000002080019c444eacd6c981baea178ef1f\t9425004")
        self.assertEquals(fps[1], "000000002000082159d404eea968b81b8ea17eef1f\t9425009")
        self.assertEquals(fps[2], "000000000000080159c404efa9689a1b8eb1faef1b\t9425012")
        self.assertEquals(fps[3], "000000000000082019c404ee8968b81b8ea1ffef1f\t9425015")
        self.assertEquals(fps[4], "000000000000088419c6b5fa8968981b8eb37aef1f\t9425018")

    def test_rdmaccs1_key44(self):
        header, output = runner.run_split("--rdmaccs/1", 7, source=MACCS_SMI)
        self.assertEqual(header[b"#type"], b"RDMACCS-OpenBabel/1")
        hex_fp, id = output[0].split(b"\t")
        self.assertFalse(hex_contains_bit(hex_fp, 43))
        self.assertEqual(output[0], b"040000000000000000000000000000000000000000\t3->bit_2")
        
    def test_rdmaccs2_key44(self):
        header, output = runner.run_split("--rdmaccs/2", 7, source=MACCS_SMI)
        self.assertEqual(header[b"#type"], b"RDMACCS-OpenBabel/2")
        hex_fp, id = output[0].split(b"\t")
        self.assertTrue(hex_contains_bit(hex_fp, 43))
        self.assertEqual(output[0], b"040000000008000000000000000000000000000000\t3->bit_2")
        #                  There's an 8 here ---^
    
    def test_both_rdmaccs_flags_gives_error(self):
        result = runner.run_exit("--rdmaccs/1 --rdmaccs/2")
        self.assertIn("argument --rdmaccs/--rdmaccs/2: not allowed with argument --rdmaccs/1", result)
        result = runner.run_exit("--rdmaccs --rdmaccs/1")
        self.assertIn("argument --rdmaccs/1: not allowed with argument --rdmaccs/--rdmaccs/2", result)
        runner.run_split("--rdmaccs --rdmaccs/2", 7, source=MACCS_SMI)

    def test_substruct(self):
        headers, fps = run_split("--substruct", 19)
        self.assertEquals(headers["#type"], "ChemFP-Substruct-OpenBabel/1")
        self.assertEquals(fps[0], "07de8d002000000000000000000000000080060000000c000000000000000080030000f8401800000030508379344c014956000055c0a44e2a0049200084e140581f041d661b10064483cb0f2925100619001393e10001007000000000008000000000000000400000000000000000\t9425004")
        self.assertEquals(fps[1], "07de0d000000000000000000000000000080460300000c0000000000000000800f0000780038000000301083f920cc09695e0800d5c0e44e6e00492190844145dc1f841d261911164d039b8f29251026b9401313e0ec01007000000000000000000000000000000000000000000000\t9425009")
        

TestFingerprintTypes = unittest2.skipIf(skip_openbabel, "OpenBabel not installed")(
    TestFingerprintTypes)


@unittest2.skipIf(skip_openbabel, "OpenBabel not installed")
class TestIdAndErrors(unittest2.TestCase, support.TestIdAndErrors):
    toolkit = "openbabel"
    _runner = runner
    
@unittest2.skipIf(skip_openbabel, "OpenBabel not installed")
class TestIO(unittest2.TestCase, support.TestIO):
    _runner = runner
    
    def test_compressed_auto(self):
        header, fps = run_split("--FP3", 19, support.PUBCHEM_SDF_GZ)
        self.assertEquals(fps[0], "0400000000b001\t9425004")
    def test_compressed_specified(self):
        header, fps = run_split("--FP3 --in sdf.gz", 19, support.PUBCHEM_SDF_GZ)
        self.assertEquals(fps[0], "0400000000b001\t9425004")
    def test_format_specified(self):
        header, fps = run_split("--FP3 --in sdf", 19, support.PUBCHEM_ANOTHER_EXT)
        self.assertEquals(fps[0], "0400000000b001\t9425004")
    def test_output(self):
        dirname = tempfile.mkdtemp(prefix="test_ob2fps")
        output_filename = os.path.join(dirname, "blah.fps")
        assert len(output_filename.split()) == 1 # ensure no whitespace
        try:
            output = runner.run("--FP3 -o " + output_filename)
            assert len(output) == 0
            with open(output_filename) as f:
                result = f.readlines()
        finally:
            shutil.rmtree(dirname)
        self.assertEquals(result[0], "#FPS1\n")
        fps = [line for line in result if not line.startswith("#")]
        self.assertEquals(len(fps), 19)
        self.assertEquals(fps[0], "0400000000b001\t9425004\n")

        
class TestMACCS(unittest2.TestCase):
    @unittest2.skipIf(not HAS_MACCS, "need MACCS support")
    def test_bitorder(self):
        result = runner.run_fps("--MACCS", 7, support.fullpath("maccs.smi"))
        # The fingerprints are constructed to test the first few bytes.
        self.assertEquals(result[0][:6], support.set_bit(2))
        self.assertEquals(result[1][:6], support.set_bit(3))
        self.assertEquals(result[2][:6], support.set_bit(4))
        self.assertEquals(result[3][:6], support.set_bit(5))
        self.assertEquals(result[4][:6], support.set_bit(9))
        ## This appears to be a bug in the OpenBabel MACCS definition
        if VERSION in ("2.2.3", "2.3.0"):
            # This is WRONG, since OB has an off-by-one error in the ring sizes
            self.assertEquals(result[5][:6], "000020")
        else:
            # which is fixed in the SVN version
            self.assertEquals(result[5][:6], support.set_bit(10))
        self.assertEquals(result[6][:6], support.set_bit(16))

TestMACCS = unittest2.skipIf(skip_openbabel, "OpenBabel not installed")(TestMACCS)

if __name__ == "__main__":
    unittest2.main()



"""sdf_reader - iterate through records in an SD file"""

# This is used by the command-line SDF reader, used when the source
# fingerprints are already encoded in fields of an SD file.

# It's also used by the RDKit parser - see the comments there.

# Copyright (c) 2010-2017 Andrew Dalke Scientific, AB (Sweden)
# See the contents of "chemfp/__init__.py" for full license details.

from __future__ import absolute_import, print_function
from . import ParseError
from . import io

__all__ = ["open_sdf", "iter_sdf_records", "iter_two_tags", "iter_title_and_tag"]

import sys
import re
import chemfp


# Do a quick check that the SD record is in the correct format
_sdf_check_pat = re.compile(r"""
.*\n     # line 1
.*\n     # line 2
.*\n     # line 3
         # next are the number of atoms and bonds. This pattern allows
         #  '  0', ' 00', '000', '00 ', '0  ' and ' 0 ', which is the
         # same as checking for field.strip().isdigit()
         # The escape '\040' is for the space character
(\040\040\d|\040\d\d|\d\d\d|\d\d\040|\d\040|\040\d\040)  # number of bonds
(\040\040\d|\040\d\d|\d\d\d|\d\d\040|\d\040|\040\d\040)  # number of bonds
        # Only space and digits are allowed before the required V2000 or V3000
[\0400-9]{28}V(2000|3000)
""", re.X)

## class FileLocation(object):
##     """A mutable instance used to track record title and position information

##     You may pass one of these to the readers to that information. I'm
##     a bit unsure about doing it this way but I liked the options of
##     passing back a 2-ple of (record, position) or records with an
##     attribute like record.position even less.

##     WARNING: the attributes '.lineno', and '._record' be set-able.
##     Passing in a non-FileLocation object may cause interoperability
##     problems in the future.
##     """
##     def __init__(self, name=None):
##         self.name = name
##         self.lineno = 1
##         self._record = None  # internal variable; it only valid enough to get the title
##     @property
##     def title(self):
##         # The title isn't needed for most cases so don't extract it unless needed
##         if self._record is None:
##             return None
##         return self._record[:self._record.find("\n")].strip()
    
##     def where(self):
##         s = "at line %s" % (self.lineno,)
##         if self.name is not None:
##             s += " of %r" % (self.name,)
##         title = self.title
##         if title:
##             s += " (title=%r)" % (title,)
##         return s

##     def info(self):
##         return dict(name=self.name, lineno=self.lineno, title=self.title)

def open_sdf(source=None, decompressor="auto", errors="strict", location=None):
    """Open an SD file and return an iterator over the SD records, as blocks of text

    source - input source. Can be None (for sys.stdin), the input filename
        as a string, or a file object.
    errors - one of "strict" (default), "log", or "ignore". Other values are experimental
    location - experimental location tracking.
    """
    # XXX Adapater until I remove the old decompressor code
    if decompressor == "auto":
        format = None
    elif decompressor == "gzip" or decompressor == "gz":
        format = "sdf.gz"
    elif decompressor == "bz2":
        format = "sdf.bz2"
    elif decompressor == "none" or decompressor == "":
        format = "sdf"
    else:
        raise AssertionError(decompressor)
    format_name, compression = io.normalize_input_format(source, format)
    error_handler = io.get_parse_error_handler(errors)
    if location is None:
        location = io.Location.from_source(source)
    location.save(record_format="sdf")
    
    fileobj, close = io.open_compressed_input(source, compression, "The SDF parser")
    return iter_sdf_records(fileobj, close, error_handler, location)

# My original implementation used a slow line-oriented parser.  That
# was decently fast, but this version, which reads a block at a time
# and works directly on those blocks, is over 3 times as fast. It's
# also a lot more complicated

def iter_sdf_records(infile, close, error_handler, location, block_size=327680):
    it = _iter_sdf_records(infile, close, error_handler, location, block_size)
    next(it)
    return it

def _iter_sdf_records(infile, close, error_handler, location, block_size):
    record = None
    recno = 0
    lineno = 1
    start_offset = end_offset = 0

    # Register how to get ahold of the different parts
    def get_lineno():
        if recno == 0:
            return 0
        return lineno

    def get_recno():
        return recno

    def get_offsets():
        return start_offset, end_offset

    def get_record():
        return record

    location.register(get_lineno=get_lineno,
                      get_recno=get_recno,
                      get_offsets=get_offsets,
                      get_record=get_record)
    yield "ready!"
    try:
        pushback_buffer = ''
        records = None
        while 1:
            if not records:
                read_data = infile.read(32768)
                if not read_data:
                    # No more data from the file. If there is something in the
                    # pushback buffer then it's an incomplete record
                    if pushback_buffer:
                        if pushback_buffer.endswith("\n$$$$"):
                            # The file is missing the terminal newline. Compensate.
                            # This will cause an extra read after the known end-of-file.
                            # (Is that a problem? It's not supposed to be one.)
                            pushback_buffer += "\n"
                        else:
                            record = None
                            if lineno == 1:
                                # No records read. Wrong format.
                                recno += 1
                                error_handler.error("Could not find a valid SD record", location)
                            else:
                                error_handler.error(
       "unexpected content at the end of the file (perhaps the last record is truncated?)",
                                    location)
                            break
                    else:
                        # We're done!
                        break
                # Join the two blocks of text. This should be enough to
                # have lots of records, so split, and the last term is
                # either a partial record or the empty string. Keep track
                # of that for use in the next go-around.
                records = (pushback_buffer + read_data).split("\n$$$$\n")
                pushback_buffer = records.pop()  # either '' or a partial record

                # It is possible though unlikely that the merged blocks of
                # text contains only an incomplete record, so this might
                # loop again. However, the joining and searching is an
                # O(n**2) operation, so I don't want to do that too often.
                # While it's possible to fix this, there should be no
                # reason to support huge records - they don't exist unless
                # you are really stretching and doing things like storing
                # images or other large data in the SD tags.
                # To prevent timing problems, don't allow huge records.
                if len(pushback_buffer) > 2000000:
                    record = None
                    recno += 1
                    error_handler.error("record is too large for this reader", location)
                    return
            else:
                # We have a set of records, one string per record. Pass them back.
                for record in records:
                    # A simple, quick check that it looks about right
                    if not _sdf_check_pat.match(record):
                        recno += 1
                        error_handler.error("incorrectly formatted record", location)
                        # If the error callback returns then just skip the record
                    else:
                        record += "\n$$$$\n"  # restore the split text
                        recno += 1
                        yield record
                    lineno += record.count("\n")
                records = None
    finally:
        location.save(lineno = get_lineno(),
                      recno = get_recno(),
                      offsets = get_offsets(),
                      record = get_record())
        if close is not None:
            try:
                close()
            finally:
                close = None


# This is complicated. I tried implementing this search with a regular
# expression but it was about 30% slower than this more direct search.

# Note: tag_substr must contain the "<" and ">"
def _find_tag_data(rec, tag_substr):
    "Return the first data line for the given tag substring, or return None"
    startpos = 0
    while 1:
        tag_start = rec.find(tag_substr, startpos)
        if tag_start == -1:
            return None

        # rfind cannot return -1 because _sdf_check_pat verified there
        # are at least 3 newlines. The +1 is safe because there's at
        # least the "<" and ">" from the tag.
        tag_line_start = rec.rfind("\n", 0, tag_start) + 1
        if rec[tag_line_start] != ">":
            # This tag is not on a data tag line. It might be the value for
            # some of the text field.
            startpos = tag_start + 1
            continue

        # This is an actual tag line. Find the start of the next line.
        # The record must end with "\n$$$$\n" so find will never return -1
        # and never return the last character position.
        next_line_start = rec.find("\n", tag_start) + 1

        # These might occur if there is no data content
        if rec[next_line_start]==">" or rec[next_line_start:next_line_start+4]=="$$$$":
            return ""

        # Otherwise, get up to the end of the line
        return rec[next_line_start:rec.find("\n", next_line_start)]

# These are not legal tag characters (while others may be against the
# SD file spec, these will break the parser)
_bad_char = re.compile(r"[<>\n\r\t\0]")

def iter_two_tags(sdf_iter, tag1, tag2):
    """Iterate over SD records to get the data lines for tag1 and tag2

    sdf_iter - an iterator which returns SD records
    tag1 - the name of the first tag
    tag2 - the name of the second tag

    Each record yields a (tag1_value, tag2_value) 2-ple. If a tag is
    present then the value is from its first data line (or the empty
    string if there is no line). If there are multiple fields with
    the same name then use the first one. If a tag does not exist,
    return None.
    """
    m = _bad_char.search(tag1)
    if m:
        raise TypeError("tag1 must not contain the character %r" % (m.group(0),))
    m = _bad_char.search(tag2)
    if m:
        raise TypeError("tag2 must not contain the character %r" % (m.group(0),))
        
    tag1_substr = "<" + tag1 + ">"
    tag2_substr = "<" + tag2 + ">"
    for rec in sdf_iter:
        yield _find_tag_data(rec, tag1_substr), _find_tag_data(rec, tag2_substr)

def iter_title_and_tag(sdf_iter, tag):
    """Iterate over SD records to get the title line and data line for the specified tag

    sdf_iter - an iterator over SD records, as text
    tag - the name of the tag value to return

    Each record yields a (title, tag_value) 2-ple where the title is
    the first line of the SD record (not a tag!) and the tag_value
    comes from the first data line for the given tag. If the tag is
    present multiple times, use the first match. If the data line is
    missing, the value is"". If the tag does not exist, the value is None.
    """
    m = _bad_char.search(tag)
    if m:
        raise TypeError("tag must not contain the character %r" % (m.group(0),))
    
    tag_substr = "<" + tag + ">"
    for rec in sdf_iter:
        yield rec[:rec.find("\n")].strip(), _find_tag_data(rec, tag_substr)

def iter_tag_and_record(sdf_iter, tag):
    m = _bad_char.search(tag)
    if m:
        raise TypeError("tag must not contain the character %r" % (m.group(0),))

    tag_substr = "<" + tag + ">"
    for rec in sdf_iter:
        yield _find_tag_data(rec, tag_substr), rec



import unittest2
import re
import os
import sys
from cStringIO import StringIO

import chemfp
import _chemfp
from chemfp import __version__ as chemfp_version

version_pattern = re.compile(r"\d+\.\d+(\.\d)?((a|b|p)\d+)?$")

class SystemTests(unittest2.TestCase):
    def test_version(self):
        m = version_pattern.match(_chemfp.version())
        self.assertNotEqual(m, None, "bad version: %s" % (_chemfp.version(),))


skip_omp = (chemfp.get_max_threads() == 1)

class OpenMPTests(unittest2.TestCase):
    def setUp(self):
        self._num_threads = chemfp.get_num_threads()
    def tearDown(self):
        chemfp.set_num_threads(self._num_threads)

    def test_num_threads_is_max_threads(self):
        self.assertEquals(chemfp.get_num_threads(), chemfp.get_max_threads())

    def test_set_to_zero(self):
        chemfp.set_num_threads(0)
        self.assertEquals(chemfp.get_num_threads(), 1)

    def test_set_to_one(self):
        chemfp.set_num_threads(1)
        self.assertEquals(chemfp.get_num_threads(), 1)

    def test_set_to_two(self):
        chemfp.set_num_threads(2)
        self.assertEquals(chemfp.get_num_threads(), 2)
    test_set_to_two = unittest2.skipIf(skip_omp, "Multiple OMP threads not available")(
        test_set_to_two)

    def test_set_to_max(self):
        chemfp.set_num_threads(chemfp.get_max_threads())
        self.assertEquals(chemfp.get_num_threads(), chemfp.get_max_threads())

    def test_set_beyond_max(self):
        chemfp.set_num_threads(chemfp.get_max_threads()+1)
        self.assertEquals(chemfp.get_num_threads(), chemfp.get_max_threads())

class SaveVersionStdout(object):
    def __init__(self, module, file):
        self.module = module
        self.prog = module.__name__
        self.file = file
        
    def __enter__(self):
        self.real_stdout = sys.stdout
        self.real_prog = self.module.parser.prog
        self.module.parser.prog = self.prog
        sys.stdout = self.file
    
    def __exit__(self, type, value, traceback):
        sys.stdout = self.real_stdout
        self.module.parser.prog = self.prog
        if issubclass(type, SystemExit):
            if value.code:
                self.file.write("I don't think you expected a %r" % (value,))
            return True
        
# Check that all of the command-line tools save a version to stdout
class CommandlineVersionTest(unittest2.TestCase):
    def check_version(self, module):
        f = StringIO()
        with SaveVersionStdout(module, f):
            module.main(args=["--version"])
        output = f.getvalue()
        expected_output = module.__name__ + " " + chemfp_version + "\n"
        self.assertEqual(output, expected_output)

    def test_simsearch(self):
        from chemfp.commandline import simsearch
        self.check_version(simsearch)
        
    def test_oe2fps(self):
        try:
            import openeye.oechem
        except ImportError:
            return
        
        from chemfp.commandline import oe2fps
        self.check_version(oe2fps)
        
    def test_rdkit2fps(self):
        try:
            import rdkit
        except ImportError:
            return
        from chemfp.commandline import rdkit2fps
        self.check_version(rdkit2fps)
        
    def test_ob2fps(self):
        try:
            import openbabel
        except ImportError:
            return
        from chemfp.commandline import ob2fps
        self.check_version(ob2fps)
        
    def test_sdf2fps(self):
        from chemfp.commandline import sdf2fps
        self.check_version(sdf2fps)
        
if __name__ == "__main__":
    unittest2.main()



from __future__ import print_function
import sys
import os
from os.path import join, exists, getmtime
import inspect
import types
import re

import chemfp
import chemfp.arena
import chemfp.bitops
import chemfp.fps_io
import chemfp.encodings

## if chemfp.has_toolkit("openbabel"):
##     from chemfp import openbabel_toolkit, openbabel_types, openbabel_patterns

## if chemfp.has_toolkit("openeye"):
##     from chemfp import openeye_toolkit, openeye_types, openeye_patterns

## if chemfp.has_toolkit("rdkit"):
##     from chemfp import rdkit_toolkit, rdkit_types, rdkit_patterns

## name = sys.argv[1]
## template_name = name + ".txt"
## rst_name = name + ".rst"
template_name, rst_name = sys.argv[1:]

from jinja2 import Template, Environment, BaseLoader, TemplateNotFound

seen_functions = {}
seen_classes = {}
seen_methods = {}

def get_obj_by_name(name):
    terms = name.split(".")
    assert terms[0] == "chemfp", name
    del terms[0]
    obj = chemfp
    for term in terms:
        obj = getattr(obj, term)

    ## if isinstance(obj, types.FunctionType):
    ##     seen_functions[name] = obj
    ## elif isinstance(obj, types.ObjectType):
    ##     seen_classes[name] = obj
    ## elif isinstance(obj, types.UnboundMethodType):
    ##     seen_methods[name] = obj
    ## else:
    ##     raise AssertionError(obj)
    seen_functions[name] = obj

    return obj

class LocalDirLoader(BaseLoader):
    def __init__(self):
        self.path = "."
    
    def get_source(self, environment, template):
        path = join(self.path, template)
        if not exists(path):
            raise TemplateNotFound(template)
        mtime = getmtime(path)
        with open(path, "rb") as f:
            source = f.read().decode('utf-8')
        return source, path, lambda: mtime == getmtime(path)
        

def docstring(name):
    obj = get_obj_by_name(name)
    doc = inspect.getdoc(obj)
    if doc is None:
        raise AssertionError("Missing docstring for " + name)
    doc = inspect.cleandoc(inspect.getdoc(obj)).rstrip("\n")
    return doc



# Remove the common part in all of the reader subclasses
def reader_header(name, classname):
    doc = docstring(name)
    lines = doc.splitlines(True)
    new_lines = []
    for line in lines:
        if "The public properties are:" in line:
            line = line.replace("The public properties are:",
                        "This class implements the :class:`chemfp.base_toolkit." + classname + "` API.")
        elif ("The public properties are" in line or
            "* metadata" in line or
            "* location" in line or
            "* closed" in line):
            continue
        new_lines.append(line)

    return "".join(new_lines)

    
def docstring_body(name):
    doc = docstring(name)
    lines = doc.splitlines(True)
    if len(lines) <= 1:
        return ""
    if lines[1].strip():
        raise AssertionError("Why is there a line 2?")
    return "".join(lines[2:])

def doublescore(value):
    return value + u"\n" + u'=' * len(value)

def singlescore(value):
    return value + u"\n" + u'-' * len(value)

def label(name):
    return "_" + name.replace(".", "_")

def get_title(first, second):
    return first or second

def indent3(text):  # indent by 3 spaces
    return "   " + text.replace("\n", "\n   ") + "\n"

def indent5(text):  # indent by 5 spaces
    return "     " + text.replace("\n", "\n     ") + "\n"


_whitespace = re.compile(r"\s+")

def call(name, depth):
    obj = get_obj_by_name(name)
    terms = name.split(".")
    ## if title is not None:
    ##     return terms[-2] + "." + terms[-1] + "() " + title
    
    if isinstance(obj, types.BuiltinFunctionType):
        return obj.__doc__.partition("\n")[0]
    
    source = inspect.getsource(obj)
    assert source, name
    try:
        start = source.index("def ") + 4
        end = source.index(":")
    except ValueError as err:
        raise ValueError(str(err) + " from " + repr(name))

    prefix_terms = terms[-1-depth:-1]
    if 0 and prefix_terms:
        text = ".".join(prefix_terms) + "."  + source[start:end]
    else:
        text = source[start:end]
    #text = terms[-2] + "." + source[start:end]
    text = text.replace("\n", "")
    return _whitespace.sub(" ", text).replace("(self, ", "(").replace("(self)", "()")


def tkname(name):
    left, mid, right = name.partition(".")
    assert left == "chemfp"
    return right

def basename(name):
    left, mid, right = name.rpartition(".")
    return right

def biname(name):  # the last two terms ("binomial name")
    terms = name.split(".")
    return terms[-2] + "." + terms[-1]

def toolkit_title(name):
    # Convert something like "chemfp.toolkit.name" into "toolkit name"
    terms = name.split(".")
    assert len(terms) == 3
    return "%s (%s)" % (terms[2], terms[1])

env = Environment(loader = LocalDirLoader())
env.filters['docstring'] = docstring
env.filters["doublescore"] = doublescore
env.filters["singlescore"] = singlescore
env.filters["label"] = label
env.filters["indent3"] = indent3
env.filters["indent5"] = indent5
env.globals["get_title"] = get_title
env.globals["call"] = call
env.globals["docstring"] = docstring
env.globals["reader_header"] = reader_header
env.globals["docstring_body"] = docstring_body
env.globals["tkname"] = tkname
env.globals["basename"] = basename
env.globals["biname"] = biname
env.globals["toolkit_title"] = toolkit_title

template = env.get_template(template_name)



chemfp_functions = []


page = template.render(
    {"chemfp": chemfp,
     "chemfp_functions": chemfp_functions})


with open(rst_name, "w") as f:
    f.write(page)
